# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module for converting to and from RDKit data structures
"""

import logging
from typing import List, Tuple, Iterable, TYPE_CHECKING

from .xyz import coordinates_to_xyz

from rdkit.Chem import AllChem as Chem

if TYPE_CHECKING:
    from rdkit.Chem import Mol, Conformer

_log = logging.getLogger(__name__)


def get_conformer(mol: "Mol", num_confs: int = 10) -> "Conformer":
    """Get lowest-energy Conformer for molecular fragment.
    If conformers don't converge, get lowest energy conformer.

    :param mol: Molecular fragment
    :type mol: Mol
    :param num_confs: Number of configurations to generate, defaults to 10.
    :type num_confs: int

    :return: Lowest-energy conformer for molecule
    :rtype: Conformer
    """
    conformers = mol.GetConformers()
    if len(conformers) < num_confs:
        # Embed num_confs conformers into molecule object
        Chem.EmbedMultipleConfs(mol, numConfs=num_confs)
        conformers = mol.GetConformers()

    res = Chem.MMFFOptimizeMoleculeConfs(mol, numThreads=0)
    # returns a list of tuples (not_converged, energy)
    if 0 in [not_converged for not_converged, _ in res]:
        idx = res.index(min(res, key=lambda x: x[1]))
        conformer = conformers[idx]
    else:
        ((not_converged, energy), conformer) = min(
            zip(res, mol.GetConformers()), key=lambda x: x[0][1])
        if not_converged:
            _log.debug(f"Solution did not converge. Lowest energy found: {energy}")
    return conformer


def _conformer_to_coordinates(
    symbols: Iterable[str],
    conformer: "Conformer"
) -> List[Tuple[str, float, float, float]]:
    """Convert conformer to element coordinates

    :param symbols: Iterable of the symbols in the conformer
    :type symbols: Iterable[str]
    :param conformer: Conformer object
    :type conformer: Conformer
    :return: List of tuples with values element name, x, y and z coordinates
    :rtype: List[Tuple[str, float, float, float]]
    """
    result = []

    for atom, symbol in enumerate(symbols):
        p = conformer.GetAtomPosition(atom)
        result.append((symbol, p.x, p.y, p.z))

    return result


def _mol_to_coordinates(
    mol: "Mol",
    num_confs: int = 10
) -> List[Tuple[str, float, float, float]]:
    """Convert molecule object to list of coordinates

    :param mol: RDKit molecule object
    :type mol: Mol
    :param num_confs: Number of molecular conformers to generate,
        defaults to 10
    :type num_confs: int, optional
    :return: List of tuples of element name and x, y, z coordinates
    :rtype: List[Tuple[str, float, float, float]]
    """
    symbols = [a.GetSymbol() for a in mol.GetAtoms()]
    conformer = get_conformer(mol=mol, num_confs=num_confs)

    return _conformer_to_coordinates(symbols=symbols, conformer=conformer)


def mol_to_xyz(mol: "Mol", num_confs: int = 10) -> str:
    """Convert molecule object to XYZ file formatted string.

    :param mol: RDKit molecule object
    :type mol: Mol
    :param num_confs: Number of molecular conformers to generate,
        defaults to 10
    :type num_confs: int, optional
    :return: XYZ file formatted string
    :rtype: str
    """
    coordinates = _mol_to_coordinates(mol=mol, num_confs=num_confs)
    number_of_atoms = mol.GetNumAtoms()
    charge = Chem.GetFormalCharge(mol)
    return coordinates_to_xyz(
        number_of_atoms=number_of_atoms,
        charge=charge,
        coordinates=coordinates
    )



# Copyright (c) Microsoft Corporation.
# Copyright (c) 2018 Jensen Group
# Licensed under the MIT License.

"""
Module for generating rdkit molobj/smiles/molecular graph from free atoms
Implementation by Jan H. Jensen, based on the paper
    Yeonjoon Kim and Woo Youn Kim
    "Universal Structure Conversion Method for Organic Molecules: From Atomic Connectivity
    to Three-Dimensional Geometry"
    Bull. Korean Chem. Soc. 2015, Vol. 36, 1769-1777
    DOI: 10.1002/bkcs.10334
"""

import itertools
import tempfile
import os
import uuid

from contextlib import contextmanager

from rdkit.Chem import rdmolops
from rdkit.Chem import rdchem
try:
    from rdkit.Chem import rdEHTTools #requires RDKit 2019.9.1 or later
except ImportError:
    rdEHTTools = None

from collections import defaultdict

import numpy as np
import networkx as nx

from rdkit import Chem
from rdkit.Chem import AllChem, rdmolops


atomic_valence = defaultdict(list)
atomic_valence[1] = [1]
atomic_valence[5] = [3,4]
atomic_valence[6] = [4]
atomic_valence[7] = [3,4]
atomic_valence[8] = [2,1,3]
atomic_valence[9] = [1]
atomic_valence[14] = [4]
atomic_valence[15] = [5,3] #[5,4,3]
atomic_valence[16] = [6,3,2] #[6,4,2]
atomic_valence[17] = [1]
atomic_valence[32] = [4]
atomic_valence[35] = [1]
atomic_valence[53] = [1]

atomic_valence_electrons = {}
atomic_valence_electrons[1] = 1
atomic_valence_electrons[5] = 3
atomic_valence_electrons[6] = 4
atomic_valence_electrons[7] = 5
atomic_valence_electrons[8] = 6
atomic_valence_electrons[9] = 7
atomic_valence_electrons[14] = 4
atomic_valence_electrons[15] = 5
atomic_valence_electrons[16] = 6
atomic_valence_electrons[17] = 7
atomic_valence_electrons[32] = 4
atomic_valence_electrons[35] = 7
atomic_valence_electrons[53] = 7


__ATOM_LIST__ = \
    ['h',  'he',
     'li', 'be', 'b',  'c',  'n',  'o',  'f',  'ne',
     'na', 'mg', 'al', 'si', 'p',  's',  'cl', 'ar',
     'k',  'ca', 'sc', 'ti', 'v ', 'cr', 'mn', 'fe', 'co', 'ni', 'cu',
     'zn', 'ga', 'ge', 'as', 'se', 'br', 'kr',
     'rb', 'sr', 'y',  'zr', 'nb', 'mo', 'tc', 'ru', 'rh', 'pd', 'ag',
     'cd', 'in', 'sn', 'sb', 'te', 'i',  'xe',
     'cs', 'ba', 'la', 'ce', 'pr', 'nd', 'pm', 'sm', 'eu', 'gd', 'tb', 'dy',
     'ho', 'er', 'tm', 'yb', 'lu', 'hf', 'ta', 'w',  're', 'os', 'ir', 'pt',
     'au', 'hg', 'tl', 'pb', 'bi', 'po', 'at', 'rn',
     'fr', 'ra', 'ac', 'th', 'pa', 'u',  'np', 'pu']


def str_atom(atom):
    """
    convert integer atom to string atom
    """
    atom = __ATOM_LIST__[atom - 1]
    return atom


def int_atom(atom):
    """
    convert str atom to integer atom
    """
    #print(atom)
    atom = atom.lower()
    return __ATOM_LIST__.index(atom) + 1


def get_UA(maxValence_list, valence_list):
    """
    """
    UA = []
    DU = []
    for i, (maxValence, valence) in enumerate(zip(maxValence_list, valence_list)):
        if not maxValence - valence > 0:
            continue
        UA.append(i)
        DU.append(maxValence - valence)
    return UA, DU


def get_atomic_charge(atom, atomic_valence_electrons, BO_valence):
    """
    """

    if atom == 1:
        charge = 1 - BO_valence
    elif atom == 5:
        charge = 3 - BO_valence
    elif atom == 15 and BO_valence == 5:
        charge = 0
    elif atom == 16 and BO_valence == 6:
        charge = 0
    else:
        charge = atomic_valence_electrons - 8 + BO_valence

    return charge


def set_atomic_charges(mol, atoms, atomic_valence_electrons,
                       BO_valences, BO_matrix, mol_charge):
    """
    """
    q = 0
    for i, atom in enumerate(atoms):
        a = mol.GetAtomWithIdx(i)
        charge = get_atomic_charge(atom, atomic_valence_electrons[atom], BO_valences[i])
        q += charge
        if atom == 6:
            number_of_single_bonds_to_C = list(BO_matrix[i, :]).count(1)
            if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                q += 1
                charge = 0
            if number_of_single_bonds_to_C == 3 and q + 1 < mol_charge:
                q += 2
                charge = 1

        if (abs(charge) > 0):
            a.SetFormalCharge(int(charge))

    return mol


def set_atomic_radicals(mol, atoms, atomic_valence_electrons, BO_valences):
    """
    The number of radical electrons = absolute atomic charge
    """
    for i, atom in enumerate(atoms):
        a = mol.GetAtomWithIdx(i)
        charge = get_atomic_charge(
            atom,
            atomic_valence_electrons[atom],
            BO_valences[i])

        if (abs(charge) > 0):
            a.SetNumRadicalElectrons(abs(int(charge)))

    return mol


def get_bonds(UA, AC):
    """
    """
    bonds = []

    for k, i in enumerate(UA):
        for j in UA[k + 1:]:
            if AC[i, j] == 1:
                bonds.append(tuple(sorted([i, j])))

    return bonds


def get_UA_pairs(UA, AC, use_graph=True):
    """
    """

    bonds = get_bonds(UA, AC)

    if len(bonds) == 0:
        return [()]

    if use_graph:
        G = nx.Graph()
        G.add_edges_from(bonds)
        UA_pairs = [list(nx.max_weight_matching(G))]
        return UA_pairs

    max_atoms_in_combo = 0
    UA_pairs = [()]
    for combo in list(itertools.combinations(bonds, int(len(UA) / 2))):
        flat_list = [item for sublist in combo for item in sublist]
        atoms_in_combo = len(set(flat_list))
        if atoms_in_combo > max_atoms_in_combo:
            max_atoms_in_combo = atoms_in_combo
            UA_pairs = [combo]

        elif atoms_in_combo == max_atoms_in_combo:
            UA_pairs.append(combo)

    return UA_pairs


def get_proto_mol(atoms):
    """
    """
    mol = Chem.MolFromSmarts("[#" + str(atoms[0]) + "]")
    rwMol = Chem.RWMol(mol)
    for i in range(1, len(atoms)):
        a = Chem.Atom(atoms[i])
        rwMol.AddAtom(a)

    mol = rwMol.GetMol()

    return mol


def chiral_stereo_check(mol):
    """
    Find and embed chiral information into the model based on the coordinates
    args:
        mol - rdkit molecule, with embeded conformer
    """
    Chem.SanitizeMol(mol)
    Chem.DetectBondStereochemistry(mol, -1)
    Chem.AssignStereochemistry(mol, flagPossibleStereoCenters=True, force=True)
    Chem.AssignAtomChiralTagsFromStructure(mol, -1)

    return


def get_atoms(mol):
    atoms = [a.GetAtomicNum() for a in mol.GetAtoms()]
    return atoms


def get_mol(smiles):
    mol = Chem.MolFromSmiles(smiles)
    Chem.Kekulize(mol, clearAromaticFlags=True)
    charge = Chem.GetFormalCharge(mol)
    mol = Chem.AddHs(mol)
    return mol


def generate_structure_from_smiles(smiles):

    # Generate a 3D structure from smiles

    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)

    status = AllChem.EmbedMolecule(mol)
    status = AllChem.UFFOptimizeMolecule(mol)

    conformer = mol.GetConformer()
    coordinates = conformer.GetPositions()
    coordinates = np.array(coordinates)

    atoms = get_atoms(mol)

    return atoms, coordinates


def read_xyz_file(filename, look_for_charge=True):
    """
    """

    atomic_symbols = []
    xyz_coordinates = []
    charge = 0
    title = ""

    with open(filename, "r") as file:
        for line_number, line in enumerate(file):
            if line_number == 0:
                num_atoms = int(line)
            elif line_number == 1:
                title = line
                if "charge=" in line:
                    charge = int(line.split("=")[1])
            else:
                atomic_symbol, x, y, z = line.split()
                atomic_symbols.append(atomic_symbol)
                xyz_coordinates.append([float(x), float(y), float(z)])

    atoms = [int_atom(atom) for atom in atomic_symbols]

    return atoms, charge, xyz_coordinates


@contextmanager
def in_temp_path(cleanup: bool = True, suffix: bool = True):
    """Contextmanager to run function in a temporary path
    This is useful for functions that generate unused output files

    :param cleanup: Cleanup the output, defaults to True
    :type cleanup: bool, optional
    :param suffix: Optional: Add random suffix to temporary path
    :type suffix: bool, optional
    """
    # Get current directory
    cur_dir = os.getcwd()
    # Create temporary path
    uid = str(uuid.uuid1()).split("-")[0] if suffix else ""
    temp_path = tempfile.TemporaryDirectory(suffix=uid)
    # Change into the temporary directory
    os.chdir(temp_path.name)
    # Run code, yield generated output path
    yield temp_path
    # Change back to original directory
    os.chdir(cur_dir)
    # Cleanup
    if cleanup:
        temp_path.cleanup()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .geometry import (
    Element,
    Geometry,
    format_geometry,
    format_geometry_from_mol,
    format_geometry_from_xyz
)

from .xyz import coordinates_to_xyz
from .rdkit_convert import (
    get_conformer,
    mol_to_xyz
)
